[comment {-*- tcl -*- doctools manpage}]
[vset libname www::websocket]
[vset cmdname www]
[manpage_begin [vset libname] n 1.0]
[copyright {2021 Schelte Bron}]
[moddesc   {WebSocket for Tcl HTTP client library}]
[titledesc {WebSocket add-on for the www Tcl library}]
[require Tcl 8.6]
[require www 2.0]
[require [vset libname] 1.0]
[see_also www]

[description]
The [package [vset libname]] package extends the [package www] package with
support for setting up websocket connections.
When the package is loaded, a new subcommand is added to the [package www]
package:

[list_begin definitions]
[call [cmd www] [method websocket] [opt "[arg option] [arg value] [arg ...]"] \
  [arg url] [opt [arg protocols]] [opt [arg extensions]]]

The [arg url] argument specifies the URL to which to connect. This should be
a URL handled by a WebSocket server. The URL should employ a ws or wss scheme.
[para]
The [arg protocols] argument can be used to indicate a list of sub-protocols
the client requests to use. The list of protocols should be ordered by
preference. See [sectref Protocols] for details.
[para]
The [arg extensions] argument may be used to request the use of one or more
websocket extensions. The [arg extensions] are defined using a dict that is
keyed on the extension name and has at least the [arg implementation] subkey.
Extension parameters may be specified using the [arg parameters] subkey.
See [sectref Extensions] for details.
[para]
The [cmd www] [method websocket] subcommand supports the following options:
[arg -auth], [arg -digest], [arg -maxredir], and [arg -timeout].
See the [cmd www] manual entry for details on these options.
[para]
Upon a successful websocket connection, the command returns a websocket object
command. See [sectref "Object Command"] below for details.
The [emph protocol] key in the return options indicates which protocol the
server selected to use, if any.
See [sectref Protocols] for details.
[list_end]

[section "Object Command"]
The websocket object command has the following subcommands:
[list_begin definitions]
[def "[term websocket] [method binary] [arg data]"]
Send binary data over the web socket. The [arg data] argument should be a
binary data string as created by e.g. the [cmd binary] [method format] or
[cmd encoding] [method convertto] commands.
[def "[term websocket] [method callback] [arg types] [arg commandPrefix]"]
Arrange for [arg commandPrefix] to be executed (with additional arguments)
whenever one of the specified message types is received on the websocket.
When the [arg commandPrefix] argument is the empty string, the callback for
the specified [arg types] is disabled. When the [arg types] argument is also
empty, all callbacks are disabled.
[para]
The base [package [vset libname]] package supports the callback types listed
below. Extensions can provide additional callback types.
[list_begin definitions]
[def "[term commandPrefix] [fun binary] [arg data]"]
This handler must be defined for a websocket connection to accept [fun binary]
messages. In the absence of a binary handler, it is considered an error to
receive a [fun binary] message. If that happens, the websocket connection is
terminated with code 1003 (unacceptable data type).
[def "[term commandPrefix] [fun close] [arg code] [arg reason]"]
If the server closes the connection using a [fun close] message without a
status code, this callback is invoked with a code of 1005 (no status code).
If the websocket connection is terminated without the server sending a
[fun close] message, this callback is invoked with a code of 1006 (connection
closed abnormally).
[def "[term commandPrefix] [fun ping] [arg data]"]
When a callback has been enabled for the [fun ping] message, it becomes the
responsibility of the callback handler to send a [fun pong] message.
[def "[term commandPrefix] [fun pong] [arg data]"]
Use when sending a [fun ping] message.
[def "[term commandPrefix] [fun text] [arg data]"]
This handler must be defined for a websocket connection to accept [fun text]
messages. In the absence of a text handler, it is considered an error to
receive a [fun text] message. If that happens, the websocket connection is
terminated with code 1003 (unacceptable data type).
[list_end]
[def "[term websocket] [method close] [opt [arg code]] [opt [arg reason]]"]
Send a [fun close] message to the server. The client expects the server to
actually drop the TCP connection.
[def "[term websocket] [method ping] [opt [arg data]]"]
Send a [fun ping] message. A [fun pong] handler should be defined to be able
to verify the response from the server.
[def "[term websocket] [method pong] [opt [arg data]]"]
Send a [fun pong] message. This should normally only be done from within a
[fun ping] handler.
[def "[term websocket] [method text] [arg string]"]
Send a [fun text] message to the server. The string will be UTF-8 encoded
before sending.
[list_end]

[section Protocols]
The list of protocols specified in the [cmd www] [method websocket] command
is offered to the WebSocket server via the Sec-WebSocket-Protocol header. If
the server agrees to using one of the offered protocols, it confirms that by
returning the selected protocol in the Sec-WebSocket-Protocol header in the
response. This information can be found in the [fun protocol] return option.
[para]
Here is an example of how to examine the return option:
[example_begin]
try {
    www websocket $url {soap mqtt}
} on ok {websock info} {
    set protocol [lb]dict get $info protocol[rb]
    if {$protocol eq "mqtt"} {
        $websock callback {binary close} mqtthandler
    } elseif {$protocol eq "soap"} {
	$websock callback {text close} soaphandler
    } else {
        $websock callback {text close} texthandler
    }
}
[example_end]
[section Extensions]
The WebSocket specification allows for extending the protocol.
The [package [vset libname]] package supports this via the [arg extensions]
argument to the [cmd www] [method websocket] command. The [arg extensions]
argument should be a dict consisting of dicts defining the implementation and
optional parameters of an extension.
[para]
For example, to request the use of the permessage-deflate extension, the
following dict may be passed as the [arg extensions] argument to the
[cmd www] [method websocket] command:
[example_begin]
permessage-deflate {
    implementation WSDeflate
    parameters {
        client_max_window_bits
        server_max_window_bits=10
    }
}
[example_end]

[para]
The extension implementation must be an TclOO class with the name specified
in the implementation subkey of the extension name. When the server accepts
the use of the extension, the implementation class is mixed into the websocket
object command.

The class can define methods that override or augment the default websocket
object methods. The class should at least provide a [method parameters] method
with a single argument. This method is invoked with the list of parameters the
server returned for the extension.
[para]
The following private methods are designed for augmenting in an extension:
[list_begin definitions]
[def [method Read]]
Read raw data from the socket. Returns the data.
[def "[method Write] [arg data]"]
Write raw data to the socket.
[def "[method Receive] [arg opcode] [arg data] [arg flags]"]
Process a received websocket message. Fragmented messages have already been
reassembled before this method is called. The [arg flags] argument is a string
of 3 bits representing the RSV1, RSV2, and RSV3 bits.
[def "[method Transmit] [arg opcode] [arg data] [opt [arg flags]]"]
Transmit a websocket message. The [arg flags] argument is expected to be a
string of up to 4 bits representing the FIN, RSV1, RSV2, and RSV3 bits.
Unused trailing bits may be omitted. So, to send a fragmented message, the
[arg flags] argument can be specified as "0" on all but the final frame.
The [arg flags] argument defaults to "1" (FIN), indicating a complete message.
[list_end]
[para]
Below is an (incomplete) outline of how an extension class for the 
permessage-deflate extension (specified in RFC7692) may be defined:
[example_begin]
oo::class create WSDeflate {
    method parameters {params} {
        my variable parameters
        set parameters $params
    }

    method Transmit {opcode data {flags 1}} {
        if {$opcode < 8} {
            set data [lb]deflate $data[rb]
        }
        next $opcode $data $flags
    }

    method Receive {opcode data flags} {
        if {$opcode < 8} {
            set data [lb]inflate $data[rb]
        }
        next $opcode $data $flags
    }
}

[example_end]


[manpage_end]
