[comment {-*- tcl -*- doctools manpage}]
[vset libname www]
[vset cmdname www]
[manpage_begin [vset libname] n 2.4]
[copyright {2020, 2021 Schelte Bron}]
[moddesc   {Tcl HTTP client library}]
[titledesc {High level HTTP client library for Tcl}]
[require Tcl 8.6]
[require Thread]
[require sqlite3]
[require tls]
[require [vset libname] 2.4]
[see_also www::proxypac]
[see_also www::websocket]
[see_also www::http2]
[description]

The [package [vset libname]] package provides high level commands for working
with the HTTP protocol, as defined in RFC 2616. The package supports the GET,
POST, HEAD, PUT, and DELETE operations of HTTP/1.1.
[para]
The [vset libname] package implements a single command, [cmd [vset cmdname]],
with several subcommands.
[usage [cmd [vset cmdname]] [arg subcommand] [opt "[arg arg] [arg arg] ..."]]
[para]
The package automatically handles compression, proxies, cookies, redirects, and
authentication so the user of the package can normally just pass the URL and
get back the page contents.
[para]
For more advanced control, meta data can be obtained, and errors cases can
be handled using the normal Tcl error handling mechanisms.

[section "Data commands"]
There are a number of subcommands that actually communicate with an HTTP
server and send and/or retrieve data. These are referred to as data commands.
The contacted HTTP server may take some time to respond. It is therefor
usually best to run data commands from inside a coroutine. Otherwise the
command will internally use the [cmd vwait] command to wait for the response.
[subsection "Command result"]
Upon successful execution of a data command, the return value is the body
received in the HTTP response. Behind the scenes, the command may have had to
perform several HTTP requests to arrive at the final response. For example,
a request for a http resource may obtain a 301 (Moved Permanently) response,
pointing to a https version of the resource. That may then produce a 401
(Unauthorized) result along with a digest authentication challenge. After the
request is repeated, including an authorization response that is constructed
based on the challenge and the credetials provided via the [option -digest]
option, a 200 (OK) result is finally obtained.
[para]
When any fatal problem is encountered while accessing the requested resource,
the command throws an error.
Possible problems include invalid command options, bad URLs, unreachable
servers, but also HTTP status codes that indicate something other than success.
For this last group of errors, the command will produce an error code of
[fun WWW] [fun CODE], followed by the group of codes (i.e. 3XX for
redirection, 4XX for client errors, 5XX for server errors), the actual
numerical status code, and the reason string.
For example, requesting a resource that does not exist on the server will
usually result in an error code [fun WWW] [fun CODE] [fun 4XX] [fun 404]
[fun "{Not Found}"].
The return value is the body received in the HTTP response.
This is usually a small html document explaining the error.
[para]
In cases where a response is received from the HTTP server, either success or
failure, it may sometimes be useful to access the meta data.
This information is available via the return options. Return options can be
obtained using the Tcl [cmd catch] and [cmd try] commands:
[example {
    catch {www get $url} data info
    # or
    try {
        www get $url
    } on ok {data info} {
        puts "Retreived [dict get $info url]"
    }
}]
[para]
The return options normally contain the following pieces of information:
[list_begin definitions]
[def url]
The URL used to retreive the returned data. This may differ from the URL
specified in the command due to redirects.
[def uri]
The resource on the server that was retreived.
[def status]
The response status information, both as the full [fmt_var line] and split
out into [fmt_var version], [fmt_var code], and [fmt_var reason] parts.
[def headers]
The response headers. All header names are returned in lowercase. Note that
this is a list of alternating header names and their values. It is not a dict,
as the same header name may appear multiple times.
[list_end]

[subsection "Common options"]
The data commands share a common set of options. These options are explained
here, and then referred to by each subcommand for brevity.
[list_begin options]
[opt_def -auth data]
Specify a username:password to send to the server for basic authorization.

[opt_def -digest cred]
Specify a username:password to use for digest authorization.

[opt_def -file file]
Send the file contents in the body of the request. The [option -file] option
may be repeated to send multiple files in a single request.

[opt_def -handler callback]
Invoke [arg callback] whenever HTTP data is available. If present, nothing else
will be done with the HTTP data.

[opt_def -headers dict]
Specify additional headers to include in the request.

[opt_def -maxredir type]
The maximum number of redirects that will be followed when retrieving a
resource. This prevents infinite loops when URLs point to themselves, either
directly or indirectly.
The default is 20.
To disable redirections, set this option to 0.
Use -1 to allow infinite redirections (not recommended).

[opt_def -multipart type]
Control whether the data will be sent as a multipart message.

[opt_def -name string]
Specify the name for the item of the following [option -file] or
[option -value] option.

[opt_def -persistent bool]
Specify whether an attempt should be made to reuse HTTP/1.1 connections. By
default this is true, as defined by RFC 2616. HTTP/1.0 connections are never
reused.

[opt_def -timeout milliseconds]
Specifies a time limit for the response to be obtained. If no response is
received within this time, the channel is closed and an error is returned to
the caller. The default is 30000 (30 seconds).

[opt_def -type mediatype]
Specify the media type of the data in the following body part(s) of the
request.

[opt_def -upgrade dict]
A map of names and their implementations for protocol upgrades that will be
offered to the server. This can be used by a client to upgrade a connection
from HTTP 1.1 to HTTP 2.0, or an HTTP or HTTPS connection into a WebSocket.
See [sectref "Protocol upgrades"] for more information.

[opt_def -value string]
Specify a fixed value for a body part. This is similar to the [arg value]
arguments, but may be used if files and values must be included in a
specific order.

[list_end]
[subsection "Request body"]
HTTP POST, PUT, and DELETE requests may have a body. There are several
possible sources of information from which the body may be constructed:
[list_begin itemized]
[item]The [arg body] argument.
[item]A single [option -file] option.
[item]Additional [option -file] options.
[item]One or more [arg key]/[arg value] pairs.
[list_end]
[para]
The following rules apply:
[list_begin enumerated]
[enum]
When the [option -multipart] option is specified, its value determines how
the body is created. When set to "form-data", [arg key]/[arg value] pairs
are included as individual body parts. Any other value, including the empty
string, will result in the [arg key]/[arg value] pairs being appended as a
query string to the url. If the [option -multipart] option is explicitly set
to "", the body will contain the body argument, if specified, or the first
file specified via a [option -file] option.
[enum]
When multiple sources are specified, the body will be created by putting all
sources into a multipart MIME data stream.
The Content-type header is set to multipart/form-data.
The value of the [option -type] option is used as the Content-type for the
part containing the data from the [arg body] argument, if present.
[enum]
If the [arg body] argument is the only specified source, the data is copied
to the message body as is.
The Content-type header is set to the value specified by the [option -type]
option. It defaults to text/plain.
[enum]
If only a single [option -file] option is specified, the contents of the file
is placed verbatim into the message body. Multiple [option -file] options
constitute multiple sources and therefor follow rule #1.
The Content-type header is set to the value specified by the [option -type]
option.
If not specified, the Content-type is inferred from the file extension.
[enum]
If only [arg key]/[arg value] pairs are specified, then the result depends on
the value specified by the [option -type] option.
For application/x-www-form-urlencoded, the data is combined into a query
string.
For multipart/form-data, the data is placed in a multipart MIME data stream.
If no [option -type] option is specified, it defaults to
application/x-www-form-urlencoded for the [method post] and [method put]
subcommands. For the [method delete] subcommand, the [arg key]/[arg value] are
appended to the URL as a query string. The message will not have a body.
[enum]
If no source for a body is specified for the [method post] or [method put]
subcommands, an empty body will be sent.
[list_end]
[para]
The following data commands are available:
[para]
[list_begin definitions]
[def "[cmd [vset cmdname]] [method get]\
  [opt "[arg option] [arg value] [arg ...]"]\
  [arg url] [opt "[arg key] [arg value] [arg ...]"]"]
[para]
Perform an HTTP GET operation. Any [arg key]/[arg value] arguments are
formatted as a query string that is appended to the URL.
[para]
The following standard options are supported by the [method get] method:
[para]
[option -auth], [option -digest], [option -handler], [option -headers],
[option -persistent], and [option -timeout].

[para]
[def "[cmd [vset cmdname]] [method post]\
  [opt "[arg option] [arg value] [arg ...]"]\
  [arg url] [opt [arg body]] [opt "[arg key] [arg value] ..."]"]
[para]
Perform an HTTP POST operation. The request must contain a body part.
See [sectref "Request body"] for more information.
[para]
The following standard options are supported by the [method post] method:
[para]
[option -auth], [option -digest], [option -file], [option -handler],
[option -headers], [option -persistent], [option -timeout], and [option -type].

[para]
[def "[cmd [vset cmdname]] [method head]\
  [opt "[arg option] [arg value] [arg ...]"]\
  [arg url] [opt "[arg key] [arg value] ..."]"]
[para]
Perform an HTTP HEAD operation. Any [arg key]/[arg value] arguments are
formatted as a query string that is appended to the URL.
[para]
The following standard options are supported by the [method head] method:
[para]
[option -auth], [option -digest], [option -handler], [option -headers],
[option -persistent], and [option -timeout].

[para]
[def "[cmd [vset cmdname]] [method put]\
  [opt "[arg option] [arg value] [arg ...]"]\
  [arg url] [opt [arg body]] [opt "[arg key] [arg value] ..."]"]
[para]
Perform an HTTP PUT operation. The request must contain a body part.
See [sectref "Request body"] for more information.
[para]
The following standard options are supported by the [method put] method:
[para]
[option -auth], [option -digest], [option -file], [option -handler],
[option -headers], [option -persistent], [option -timeout], and [option -type].

[para]
[def "[cmd [vset cmdname]] [method delete]\
  [opt "[arg option] [arg value] [arg ...]"]\
  [arg url] [opt [arg body]] [opt "[arg key] [arg value] ..."]"]
[para]
Perform an HTTP DELETE operation. The request may contain a body part.
See [sectref "Request body"] for more information.
[para]
The following standard options are supported by the [method delete] method:
[para]
[option -auth], [option -digest], [option -file], [option -handler],
[option -headers], [option -persistent], [option -timeout], and [option -type].
[list_end]
[subsection "Control commands"]
[list_begin definitions]
[def "[cmd [vset cmdname]] [method log] [arg prefix]"]
[para]
The [cmd [vset cmdname]] [method log] command configures a command prefix for
handling log messages.
Whenever the library has something to report, the log message is appended as
an extra argument to [arg prefix] and the resulting command is executed.
[para]
Example: Print all log messages to standard output:
[example "[vset cmdname] log puts"]
[para]
[def "[cmd [vset cmdname]] [method configure]\
  [opt "[option -accept] [arg mimetypes]"]\
  [opt "[option -useragent] [arg string]"]\
  [opt "[option -proxy] [arg prefix]"]\
  [opt "[option -pipeline] [arg boolean]"]\
  [opt "[option -socketcmd] [arg prefix]"]"]
[para]
Configure various aspects of the behavior of the package. The following options
are supported:
[list_begin options]
[opt_def -accept mimetypes]
The Accept header of the request. The default is */*, which means that all
types of documents are accepted. Otherwise you can supply a comma-separated
list of mime type patterns that you are willing to receive.
For example, "image/gif, image/jpeg, text/*".

[opt_def -useragent string]
The value of the User-Agent header in the HTTP request. 
The default is "[emph "Tcl-www/1.0"]".

[opt_def -proxy prefix]
Control the use of a proxy for sending HTTP requests.
See [sectref Proxies] below for more information.

[opt_def -pipeline boolean]
Control the use of pipelining.
Pipelining means that multiple HTTP are sent on a single TCP connection
without waiting for the corresponding responses.
Disabled by default.

[opt_def -socketcmd prefix]
Define a command prefix to use when creating a TCP socket. When necessary, the
command is executed in a separate thread with three additional arguments:
[option -async], [arg host], and [arg port]. Because the command will run in a
separate thread, it has to be self-contained and only rely on standard Tcl
commands.
For example, to use the [cmd socket] command of the [package iocp] package,
the following code could be used:
[example {
    www configure -socketcmd [list apply [list {path args} {
        set ::auto_path $path
        package require iocp_inet
        tailcall socket {*}$args
    }] $auto_path]
}]

[list_end]
[para]
[def "[cmd [vset cmdname]] [method cookies] [arg subcommand]\
  [opt "[arg arg] [arg arg] ..."]"]
[para]
Manage the cookies related to a url. The [method cookies] subcommand provides
the following methods:
[list_begin definitions]
[def "[cmd [vset cmdname]] [method cookies] [method delete]\
  [arg url] [opt "[arg cookie] ..."]"]
Delete zero or more cookies that have been stored for the specified [arg url].
If no [arg cookie] is specified, all cookies for the [arg url] are deleted.

[def "[cmd [vset cmdname]] [method cookies] [method get] [arg url]"]
Return all cookies applicable for the specified [arg url].
The command returns a list of alternating cookie names and their value.

[def "[cmd [vset cmdname]] [method cookies] [method store]\
  [arg url] [opt "[arg cookiespec] ..."]"]
Add or overwrite a cookie. The [arg cookiespec] is a string defining the
cookie name, value and attributes as it might appear in an http Set-Cookie
header. For example:

[example {id=a3fWa; Expires=Wed, 21 Oct 2015 07:28:00 GMT; HttpOnly}]
[list_end]
[para]

[def "[cmd [vset cmdname]] [method register]\
  [arg scheme] [arg port] [arg prefix] [opt [arg secure]]"]
[para]
Register a custom HTTP scheme by specifying the scheme name, default port and
the command for creating the Tcl channel.
The optional [arg secure] argument should be a boolean indicating whether the
channel may be considered secure. This influences sending of secure cookies.
[para]
[def "[cmd [vset cmdname]] [method certify] [arg cainfo] [opt [arg prefix]]"]
[para]
Specify the location of the trusted CA certificates. The location may be a
directory containing individual CA certificates, or one or more CA certificates
concatenated into a single file.
[para]
Invoking this command activates the verification of the certificate returned
by the server against a set of trusted certificate authorities.
When [arg prefix] is not specified, standard checks are performed, unless an
empty string is passed for [arg cainfo]. In that case, any certificate is
accepted.
[para]
For finer grained control, a command [arg prefix] can be specified.
This command will be called at several points during the OpenSSL handshake
and may reject or accept the certificate at each point by returning
a false or true boolean value, respectively.
If the callback does not return a valid boolean value, the result is
determined using the standard checks.
The callback is invoked by appending two arguments to the provided prefix:
[list_begin itemized]
[item]A [arg depth] argument, which is an integer representing the current
depth on the certificate chain, with 0 as the subject certificate and higher
values denoting progressively more indirect issuer certificates.
[item]A [arg cert] argument, containing a list of key-value pairs providing
details about the received certificate.
[list_end]

[para]
[def "[cmd [vset cmdname]] [method cookiedb] [arg database]"]
[para]
Specify the location of the sqlite3 database to use for storing permanent
cookies. The same database may be used by multiple applications using the
[vset libname] package to share cookies.
If no database is configured, the package uses a temporary database, which
gets deleted when the application terminates.

[para]
[def "[cmd [vset cmdname]] [method header] [arg option] [opt "[arg arg] ..."]"]
[para]
This command provides several operations for working with a list of http
headers such as have been returned by the different data commands.

[list_begin definitions]
[def "[cmd [vset cmdname]] [method header] [method add] \
  [arg headerlistvar] [arg name] [opt [arg -nocase]] [arg value] [opt ...] "]
Add one or more values to a header, if they are not alread present.
The [arg -nocase] option makes the compare operation case insensitive.
[def "[cmd [vset cmdname]] [method header] [method append] \
  [arg headerlistvar] [arg name] [opt [arg value]] [opt ...] "]
Set a new value for a header in addition to any existing values.
[def "[cmd [vset cmdname]] [method header] [method exists] \
  [arg headerlist] [arg name]"]
Check if a header with the specified name exists.
[def "[cmd [vset cmdname]] [method header] [method get] \
  [arg headerlist] [arg name] [opt [arg index]] [opt [arg -lowercase]]"]
Return the value of the requested header, if any. By default all entries are
joined together, separated with a comma and a space. The resulting string is
returned.
If an index is specified, that is taken as an indication that the header
value is defined as a comma-separated list. In that case, a Tcl list is
constructed from the individual elements of all entries. The requested index
from the resulting list is returned.
The special index "all" causes the complete list to be returned.
When the [arg -lowercase] option is specified, all values are converted to
lower case.
[def "[cmd [vset cmdname]] [method header] [method replace] \
  [arg headerlistvar] [arg name] [opt [arg value]] [opt ...] "]
Set a new value for a header replacing all existing entries.
Multiple values are joined together into a comma-separated list.
If no values are specified, all entries for the header are removed.
[list_end]
[list_end]

[section Proxies]
The [vset libname] package supports the use of a proxy to access resources.
The desired method can be configured using the [option -proxy] option of
the [cmd [vset cmdname]] [method configure] subcommand.
The [option -proxy] option takes a command prefix that will be called
before a URL is fetched. The command is invoked with two additional arguments:
[list_begin definitions]
[def [arg url]]
The URL for the request. This does not include the query string, if any.
[def [arg host]]
The host-name derived from the URL (i.e. the string between '://' and the
first subsequent ':' or '/').
[list_end]
[para]
The command should return a list containing one or more of the following:
[para]
[list_begin definitions]
[def DIRECT]
Do not use a proxy for this URL.
[def "PROXY [arg server]"]
Use the indicated proxy server via HTTP.
[def "SOCKS [arg server]"]
Use the indicated SOCKS4 proxy server.
[def "HTTP [arg server]"]
Same as PROXY.
[def "HTTPS [arg server]"]
Use the indicated proxy server via HTTPS.
[def "SOCKS4 [arg server]"]
Same as SOCKS.
[def "SOCKS5 [arg server]"]
Use the indicated SOCKS5 proxy server.
[list_end]
[para]
The [arg server] argument may be specified as "host[lb]:port[rb]". If the
port part is omitted, it defaults to 8080 for HTTP(S) proxies, and 1080 for
SOCKS([lb]45[rb]) proxies.

[subsection "Proxy handlers"]
The [vset libname] package provides a number of predefined proxy options:
[list_begin commands]
[cmd_def [method noproxy]]
Do not use a proxy for any URL.
[cmd_def [method defaultproxy]]
Use proxies based on the HTTP_PROXY, HTTPS_PROXY, and NO_PROXY environment
variables. The variable names are not case sensitive.
[cmd_def "[method httpproxy] [arg server]"]
Use the indicated proxy server via HTTP.
[cmd_def "[method httpsproxy] [arg server]"]
Use the indicated proxy server via HTTPS.
[cmd_def "[method socksproxy] [arg server]"]
Use the indicated SOCKS4 proxy server.
[cmd_def "[method socks4proxy] [arg server]"]
Use the indicated SOCKS4 proxy server.
[cmd_def "[method socks5proxy] [arg server]"]
Use the indicated SOCKS5 proxy server.
[list_end]
[para]
An additional proxy option is available via the [package www::proxypac]
package. This option allows proxies to be selected based on a proxy
auto-configuration (PAC) file. See the
[sectref-external "[package www::proxypac] manual page"] for more information.

[section "Protocol upgrades"]
HTTP/1.1 connections may be transformed to a different protocol using the
protocol upgrade mechanism. The client invites the server to switch to one
or more alternative protocols by providing a list of protocol names in
descending order of preference.
The [package [vset libname]] package supports this mechanism via the
[arg -upgrade] option. The option takes a dict that maps protocol names to
their implementation. The offered protocols may require additional arguments.
These can be provided using the [arg -headers] option.
[para]
The protocol implementation must be provided in the form of an oo::class.
If the server decides to accept the upgrade, the oo::class is mixed into the
object command that was created for the HTTP connection.
The protocol implementation class must define at least a [method Startup]
method that accepts a single argument. This method is invoked immediately
after the class is added to the connection object. A dict containing a
key-value list of the headers from the server's response is passed to the
[method Startup] method.
[para]
The [package [vset libname]] package comes with upgrade implementations for
WebSockets and HTTP/2. These can be utilized by loading their respective
packages:
[example_begin]
package require [package www::websocket]
package require [package www::http2]
[example_end]
See the manual entries for the packages for details of their use.
[manpage_end]
